import datetime
import logging
import os

from langchain.docstore.document import Document
from scrapy.exceptions import DropItem

from langsearch.pipelines.base import BasePipeline
from langsearch.pipelines.common.mixins.weaviatedb import WeaviateMixin


logger = logging.getLogger(__name__)


class BaseSimpleIndexPipeline(BasePipeline):
    INPUTS = {
        "url": "url",
        "sections": "sections",
        "changed": "changed"
    }
    CLASS_SCHEMA = {
        "class": "Section",
        "vectorizer": "text2vec-transformers",
        "moduleConfig": {
            "text2vec-transformers": {
                "vectorizeClassName": False,
            }
        },
        "properties": [
            {
                "name": "url",
                "description": "The URL containing the section",
                "dataType": ["string"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": True,
                    }
                },
            },
            {
                "name": "section",
                "description": "Section text",
                "dataType": ["text"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": False,
                        "vectorizePropertyName": False
                    }
                },
            },
            {
                "name": "last_seen",
                "description": "When this URL was last seen in the crawling process",
                "dataType": ["date"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": True,
                    }
                },
            }
        ],
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        class_schema = self.__class__.get_setting_from_partial_key(os.environ, "CLASS_SCHEMA")
        if isinstance(class_schema, str):
            class_schema = self.get_params_from_file(class_schema)
        self.class_schema = class_schema
        self.class_name = self.class_schema["class"]
        if not self.weaviate.class_exists(self.class_name):
            self.weaviate.create_class(self.class_schema)

    def create_or_change(self):
        self.weaviate.client.batch.delete_objects(
            class_name=self.class_name,
            where={
                "path": ["url"],
                "operator": "Equal",
                "valueString": self.url
            }
        )
        self.weaviate.add(
            class_name=self.class_name,
            data=[
                {
                    "url": self.url,
                    "section": section,
                    "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                } for section in self.sections
            ]
        )

    def apply(self, item, spider):
        if not hasattr(self, "url"):
            return item
        if not hasattr(self, "sections"):
            return item
        try:
            # TODO: Check if sections for this URL exists; otherwise create them
            if hasattr(self, "changed") and not self.changed:
                self.weaviate.update_property_with_current_datetime(
                    class_name=self.class_name,
                    where_filter={
                        "path": ["url"],
                        "operator": "Equal",
                        "valueString": self.url
                    },
                    property_name="last_seen"
                )
            else:
                self.create_or_change()
            return item
        except:
            message = f"Error while processing item with URL {self.url}"
            logger.exception(message)
            raise DropItem(message)

    def get_similar_sections(self, text, **kwargs):
        result = self.weaviate.get_similar(self.class_name, text, ["url", "section", "last_seen"], **kwargs)
        return [Document(page_content=item["section"], metadata={"source": item["url"]})
                for item in result
                ]


class SimpleIndexPipeline(BaseSimpleIndexPipeline, WeaviateMixin):
    pass



class DryRunPipeline:
    def __init__(self, filepath):
        self.filepath = filepath
        self.file = None

    @classmethod
    def from_crawler(cls, crawler):
        filepath = crawler.settings.get("LANGSEARCH_DRY_RUN_PIPELINE_FILEPATH", "dry_run_results.txt")
        return cls(filepath)

    def open_spider(self, spider):
        self.file = open(self.filepath, "w")

    def close_spider(self, spider):
        self.file.close()

    def process_item(self, item, spider):
        self.file.write(f"{item['response'].url}\n")




from langsearch.pipelines.common.storeitem import StoreItemPipeline
from langsearch.pipelines.types.image.resize import ResizeImagePipeline
from langsearch.pipelines.types.image.imageindex import ImageIndexPipeline
from langsearch.pipelines.types.enumerations import ItemType


class GenericImagePipeline:
    """
    Built-in pipelines should always use priorities between 400 and 600.
    """
    ITEM_TYPE = ItemType.IMAGE

    ITEM_PIPELINES = {
        ResizeImagePipeline: 400,
        StoreItemPipeline: 410,
        ImageIndexPipeline: 420
    }

    RESIZE_IMAGE_PIPELINE_INPUTS = {
        "body": lambda item: getattr(item["response"], "body"),
        "url": lambda item: getattr(item["response"], "url")
    }

    STORE_ITEM_PIPELINE_INPUTS = {
        "url": lambda item: getattr(item["response"], "url")
    }

    IMAGE_INDEX_PIPELINE_INPUTS = {
        "body": ResizeImagePipeline.RESIZED_BYTES,
        "url": lambda item: getattr(item["response"], "url"),
        "changed": StoreItemPipeline.CHANGED
    }

    PIPELINE_INPUTS = {
        ResizeImagePipeline: RESIZE_IMAGE_PIPELINE_INPUTS,
        StoreItemPipeline: STORE_ITEM_PIPELINE_INPUTS,
        ImageIndexPipeline: IMAGE_INDEX_PIPELINE_INPUTS
    }



import datetime
from io import BytesIO, BufferedReader
import logging
import os

from scrapy.exceptions import DropItem
from weaviate.util import image_encoder_b64, image_decoder_b64

from langsearch.pipelines.base import BasePipeline
from langsearch.pipelines.common.mixins.weaviatedb import WeaviateMixin


logger = logging.getLogger(__name__)


class BaseImageIndexPipeline(BasePipeline):
    INPUTS = {
        "url": "url",
        "body": "body",
        "changed": "changed"
    }
    CLASS_SCHEMA = {
        "class": "Image",
        "vectorizer": "multi2vec-clip",
        "moduleConfig": {
            "multi2vec-clip": {
                "imageFields": [
                    "image"
                ],
            }
        },
        "properties": [
            {
                "name": "url",
                "description": "The URL of the image",
                "dataType": ["string"],
            },
            {
                "name": "image",
                "description": "Base 64 encoded image",
                "dataType": ["blob"],
            },
            {
                "name": "last_seen",
                "description": "When this URL was last seen in the crawling process",
                "dataType": ["date"],
            }
        ],
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        class_schema = self.__class__.get_setting_from_partial_key(os.environ, "CLASS_SCHEMA")
        if isinstance(class_schema, str):
            class_schema = self.get_params_from_file(class_schema)
        self.class_schema = class_schema
        self.class_name = self.class_schema["class"]
        if not self.weaviate.class_exists(self.class_name):
            self.weaviate.create_class(self.class_schema)

    def apply(self, item, spider):
        if not hasattr(self, "url"):
            return item
        if not hasattr(self, "body"):
            return item
        try:
            if hasattr(self, "changed") and not self.changed:
                self.weaviate.update_property_with_current_datetime(
                    class_name=self.class_name,
                    where_filter={
                        "path": ["url"],
                        "operator": "Equal",
                        "valueString": self.url
                    },
                    property_name="last_seen"
                )
            else:
                with BytesIO(self.body) as f:
                    base_64_encoded_image = image_encoder_b64(BufferedReader(f))
                    self.weaviate.client.data_object.create(
                        class_name=self.class_name,
                        data_object={
                            "url": self.url,
                            "image": base_64_encoded_image,
                            "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                        }
                    )
        except:
            message = f"Error while processing item with URL {self.url}"
            logger.exception(message)
            raise DropItem(message)

    def get_image_bytes(self, base_64_encoded_image):
        return image_decoder_b64(base_64_encoded_image)

    def get_similar_images_from_text(self, text, limit=4):
        return self.weaviate.get_similar(self.class_name, text, ["url", "image", "last_seen"], limit=limit)

    def get_similar_images_from_image(self, image, top=4):
        return self.weaviate.get_near_image(self.class_name, image, ["url", "image", "last_seen"], top)


class ImageIndexPipeline(BaseImageIndexPipeline, WeaviateMixin):
    pass



import logging

from langdetect import detect
from scrapy.exceptions import DropItem

from langsearch.pipelines.base import BasePipeline
from langsearch.exceptions import SettingsError


logger = logging.getLogger(__name__)


class LanguageFilterPipeline(BasePipeline):
    INPUTS = {
        "text": "text",
        "url": "url"
    }
    ALLOWED_LANGUAGES = None
    LANGUAGE = "language_filter_pipeline_language"

    def __init__(self, allowed_languages, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.allowed_languages = allowed_languages

    @classmethod
    def from_crawler(cls, crawler):
        allowed_languages = cls.get_setting_from_partial_key(crawler.settings, "ALLOWED_LANGUAGES")
        if not isinstance(allowed_languages, list) and allowed_languages is not None:
            raise SettingsError(
                f"setting with partial key ALLOWED_LANGUAGES of class {cls} must be a list or None, "
                f"got {type(allowed_languages)}"
            )
        return cls(allowed_languages)

    def apply(self, item, spider):
        if self.allowed_languages is None:
            return item
        else:
            if not hasattr(self, "text") or self.text is None:
                return item
            if not hasattr(self, "url"):
                return item
            try:
                language = detect(self.text)
            except:
                message = f"Language detection failed for url {self.url}"
                logger.exception(message)
                raise DropItem(message)
            else:
                if language in self.allowed_languages:
                    item[self.LANGUAGE] = language
                    return item
                else:
                    message = f"Language {language} not in allowed languages {self.allowed_languages} for url {self.url}"
                    logger.info(message)
                    raise DropItem(message)



import logging

from scrapy.exceptions import DropItem
from trafilatura import extract

from langsearch.exceptions import SettingsError
from langsearch.pipelines.base import BasePipeline

logger = logging.getLogger(__name__)


class TrafilaturaPipeline(BasePipeline):
    INPUTS = {
        "html": "html",
        "url": "url"
    }
    TEXT_WITHOUT_BP = "trafilatura_pipeline_text_without_bp"
    EXTRACT_ARGUMENTS = {}

    def __init__(self, trafilatura_extract_arguments, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trafilatura_extract_arguments = trafilatura_extract_arguments

    @classmethod
    def from_crawler(cls, crawler):
        trafilatura_extract_arguments = cls.get_setting_from_partial_key(crawler.settings, "EXTRACT_ARGUMENTS")
        if not isinstance(trafilatura_extract_arguments, dict):
            raise SettingsError(
                f"setting with partial key EXTRACT_ARGUMENTS of class {cls} must be a dict, "
                f"got {type(trafilatura_extract_arguments)}"
            )
        return cls(trafilatura_extract_arguments)

    def apply(self, item, spider):
        if not hasattr(self, "html") or self.html is None:
            return item
        if not hasattr(self, "url"):
            return item
        args = {**self.trafilatura_extract_arguments, "output_format": "text"}
        try:
            text_without_bp = extract(self.html, **args)
        except:
            message = f"Trafilatura failed to extract text for url {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            if isinstance(text_without_bp, str):
                item[self.TEXT_WITHOUT_BP] = text_without_bp
                return item
            else:
                message = f"Trafilatura extraction returned a non-string for url {self.url}"
                logger.warning(message)
                raise DropItem(message)



from langsearch.pipelines.detect import DetectItemTypePipeline
from langsearch.pipelines.dryrun import DryRunPipeline
from langsearch.pipelines.types.assemble import assemble
from langsearch.pipelines.types.html.htmlpipeline import GenericHTMLPipeline



import graphlib


def assemble(*built_in_pipelines):
    graph = {}
    pipeline_inputs = {}
    for built_in_pipeline in built_in_pipelines:
        sorted_pipelines = [pipeline for (pipeline, priority) in
                            sorted(built_in_pipeline.ITEM_PIPELINES.items(), key=lambda x: x[1])
                            ]
        for i, pipeline in enumerate(sorted_pipelines):
            predecessors = sorted_pipelines[:i]
            if len(predecessors) == 0:
                if pipeline not in graph:
                    graph[pipeline] = set()
            else:
                try:
                    graph[pipeline].update(predecessors)
                except KeyError:
                    graph[pipeline] = set(predecessors)
            for key, value in built_in_pipeline.PIPELINE_INPUTS[pipeline].items():
                try:
                    pipeline_inputs[pipeline][key][built_in_pipeline.ITEM_TYPE] = value
                except KeyError:
                    try:
                        pipeline_inputs[pipeline][key] = {built_in_pipeline.ITEM_TYPE: value}
                    except KeyError:
                        pipeline_inputs[pipeline] = {key: {built_in_pipeline.ITEM_TYPE: value}}
    ts = graphlib.TopologicalSorter(graph)
    try:
        final_order = list(ts.static_order())
    except graphlib.CycleError:
        raise RuntimeError("The built-in pipelines have a circular dependency")
    start = 400
    final_pipeline = {}
    for pipeline in final_order:
        pipeline.INPUTS = pipeline_inputs[pipeline]
        final_pipeline[pipeline] = start
        start += 1
        if start == 601:
            raise RuntimeError("The built-in pipelines have too many items")
    return final_pipeline



import datetime
import logging
import re
import os

from langchain.chains import LLMChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.docstore.document import Document
from langchain.llms import OpenAIChat
from langchain.prompts import PromptTemplate
from scrapy.exceptions import DropItem

from langsearch.pipelines.common.index import BaseSimpleIndexPipeline
from langsearch.pipelines.common.mixins.summary import RecursiveReduceSummaryMixin
from langsearch.pipelines.common.mixins.weaviatedb import WeaviateMixin
from langsearch.utils import openai_length_function


logger = logging.getLogger(__name__)


PROMPT_TEMPLATE = """You are trying to find links that might contain the answer to the question: {question}

You have a few links, but you can't view all the information contained under the link. You only have access to a concise and incomplete summary of the information contained in those links. Therefore, the summaries may not contain the answer to the question directly. The links themselves contain a lot more information than the summary. You need to decide which links to investigate further, i.e view their full content.

{context}

For which links would you fetch the full content to see if they contain the answer to the following question: {question}

Remember, the summaries may not contain the answer to the question directly, because they are incomplete. The links themselves contain a lot more information than the summary.

Please provide a list of all those links to investigate further.

List of links:
"""

PROMPT = PromptTemplate(template=PROMPT_TEMPLATE, input_variables=["question", "context"])


class BaseSummaryIndexPipeline(BaseSimpleIndexPipeline):
    INPUTS = {
        "url": "url",
        "sections": "sections",
        "changed": "changed",
        "summarizer_input": "summarizer_input",
    }
    SUMMARY_CLASS_SCHEMA = {
        "class": "Summary",
        "vectorizer": "text2vec-transformers",
        "moduleConfig": {
            "text2vec-transformers": {
                "vectorizeClassName": False,
            }
        },
        "properties": [
            {
                "name": "url",
                "description": "The URL of the document",
                "dataType": ["string"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": True,
                    }
                },
            },
            {
                "name": "summary",
                "description": "Summary of the page",
                "dataType": ["text"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": False,
                        "vectorizePropertyName": False
                    }
                },
            },
            {
                "name": "last_seen",
                "description": "When this URL was last seen in the crawling process",
                "dataType": ["date"],
                "moduleConfig": {
                    "text2vec-transformers": {
                        "skip": True,
                    }
                },
            }
        ],
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        summary_class_schema = self.__class__.get_setting_from_partial_key(os.environ, "SUMMARY_CLASS_SCHEMA")
        if isinstance(summary_class_schema, str):
            summary_class_schema = self.get_params_from_file(summary_class_schema)
        self.summary_class_schema = summary_class_schema
        self.summary_class_name = self.summary_class_schema["class"]
        if not self.weaviate.class_exists(self.summary_class_name):
            self.weaviate.create_class(self.summary_class_schema)

    def get_all_summaries_for_url(self):
        where_filter = {
            "path": ["url"],
            "operator": "Equal",
            "valueString": self.url
        }
        result = (
            self.weaviate.client.query
            .get(self.summary_class_name)
            .with_additional(["id"])
            .with_where(where_filter)
            .do()
        )
        return result["data"]["Get"][self.summary_class_name]

    def apply(self, item, spider):
        if not hasattr(self, "url"):
            return item
        if not hasattr(self, "summarizer_input"):
            return item
        try:
            data = self.get_all_summaries_for_url()
            count = len(data)
            if count > 1:
                message = (
                        f"Found {count} data obj for unique property URL {self.url}"
                        f"in class {self.summary_class_name}"
                )
                logger.warning(message)
                raise RuntimeError(message)
            elif count == 0:
                self.weaviate.client.data_object.create(
                    class_name=self.summary_class_name,
                    data_object={
                        "url": self.url,
                        "summary": self.summarize(**self.summarizer_input),
                        "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                    }
                )
            else:
                if hasattr(self, "changed") and not self.changed:
                    self.weaviate.update_property_with_current_datetime(
                        class_name=self.summary_class_name,
                        where_filter={
                            "path": ["url"],
                            "operator": "Equal",
                            "valueString": self.url
                        },
                        property_name="last_seen"
                    )
                else:
                    unique_id = data[0]["_additional"]["id"]
                    self.weaviate.client.data_object.delete(
                        class_name=self.summary_class_name,
                        uuid=unique_id
                    )
                    self.weaviate.client.data_object.create(
                        class_name=self.summary_class_name,
                        data_object={
                            "url": self.url,
                            "summary": self.summarize(**self.summarizer_input),
                            "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                        }
                    )
        except:
            message = f"Error while storing summary for item with URL {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            super().apply(item, spider)

    @staticmethod
    def extract_links(text):
        return re.findall("(?:http|file)s?://[^\s]+", text)

    def get_similar_summaries(self, text, **kwargs):
        result = self.weaviate.get_similar(self.summary_class_name, text, ["url", "summary", "last_seen"], **kwargs)
        return [Document(page_content=item["summary"], metadata={"source": item["url"]})
                for item in result
                ]

    def get_similar_summaries_using_llm(
            self,
            text,
            llm=OpenAIChat(temperature=0, max_tokens=256),
            max_tokens=256,  # This should match the max_tokens of the LLM,
            max_context_size=4096,
            length_function=openai_length_function,
            prompt=PROMPT,
            document_variable_name="context",
            document_prompt_template="LINK: {source}\nSUMMARY: {page_content}",
            **kwargs
            ):
        prompt_length = length_function(prompt.format(**{input_var: "" for input_var in prompt.input_variables}))
        document_prompt = PromptTemplate(
            input_variables=["page_content", "source"],
            template=document_prompt_template
        )
        llm_chain = LLMChain(prompt=prompt, llm=llm)
        stuff_chain = StuffDocumentsChain(
            llm_chain=llm_chain,
            document_prompt=document_prompt,
            document_variable_name=document_variable_name
        )

        summaries = self.get_similar_summaries(text, **kwargs)
        token_count = 0
        remaining = max_context_size - prompt_length - max_tokens - 50  # -50 for safety
        matching_summaries = []
        docs = []
        for summary in summaries:
            token_count += length_function(f"LINK: {summary.metadata['source']}\nSUMMARY: {summary.page_content}\n\n")
            if token_count > remaining:
                answer = stuff_chain.run(input_documents=docs, question=text)
                logger.debug(f"Summary select answer: {answer}")
                extracted_links = self.extract_links(answer)
                matching_summaries.extend([doc for doc in docs if doc.metadata["source"] in extracted_links])
                token_count = 0
                docs = [summary]
            else:
                docs.append(summary)
        answer = stuff_chain.run(input_documents=docs, question=text)
        logger.debug(f"Summary select answer: {answer}")
        extracted_links = self.extract_links(answer)
        matching_summaries.extend([doc for doc in docs if doc.metadata["source"] in extracted_links])
        return matching_summaries


class SummaryIndexPipeline(BaseSummaryIndexPipeline, WeaviateMixin, RecursiveReduceSummaryMixin):
    pass



import copy
import logging
import os

from langchain.chains.summarize import load_summarize_chain
from langchain.docstore.document import Document
from langchain.llms.loading import load_llm_from_config
from langchain.prompts import PromptTemplate


from langsearch.exceptions import SettingsError
from langsearch.utils import openai_length_function


logger = logging.getLogger(__name__)


PROMPT = """{text}

Tl;dr
"""


class RecursiveReduceSummaryMixin:
    SUMMARY_SERIALIZED_LLM = {
        "_type": "openai",
        "temperature": 0,
        "model_name": "gpt-3.5-turbo",
        "max_tokens": 512  # This should match SUMMARY_MAX_LENGTH
    }
    SUMMARY_LLM_MAX_CONTEXT_SIZE = 4096
    SUMMARY_LLM_LENGTH_FUNCTION = openai_length_function
    SUMMARY_MAX_LENGTH = 512
    SUMMARY_PROMPT = PROMPT

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # TODO: Use from_crawler() for these settings
        summary_serialized_llm = self.__class__.get_setting_from_partial_key(os.environ, "SUMMARY_SERIALIZED_LLM")
        if isinstance(summary_serialized_llm, str):
            summary_serialized_llm = self.load_params_from_file(summary_serialized_llm)
        # load_llm_from_config() modifies the dict, so we need to make a copy
        self.summary_llm = load_llm_from_config(copy.copy(summary_serialized_llm))
        summary_llm_max_context_size = self.__class__.get_setting_from_partial_key(
            os.environ, "SUMMARY_LLM_MAX_CONTEXT_SIZE"
        )
        if isinstance(summary_llm_max_context_size, str):
            try:
                summary_llm_max_context_size = int(summary_llm_max_context_size)
            except ValueError:
                raise SettingsError(
                    f"setting with partial key SUMMARY_LLM_CONTEXT_SIZE of class {self.__class__}"
                    f"must be convertible to int, but got '{summary_llm_max_context_size}'"
                )
        self.summary_llm_max_context_size = summary_llm_max_context_size
        summary_max_length = self.__class__.get_setting_from_partial_key(os.environ, "SUMMARY_MAX_LENGTH")
        if isinstance(summary_max_length, str):
            try:
                summary_max_length = int(summary_max_length)
            except ValueError:
                raise SettingsError(
                    f"setting with partial key SUMMARY_MAX_LENGTH of class {self.__class__}"
                    f"must be convertible to int, but got '{summary_max_length}'"
                )
        self.summary_max_length = summary_max_length
        summary_llm_length_function = self.__class__.get_setting_from_partial_key(
            os.environ, "SUMMARY_LLM_LENGTH_FUNCTION"
        )
        if isinstance(summary_llm_length_function, str):
            summary_llm_length_function = self.get_from_dotted(summary_llm_length_function)
        self.summary_llm_length_function = summary_llm_length_function
        summary_prompt = self.__class__.get_setting_from_partial_key(os.environ, "SUMMARY_PROMPT")
        summary_prompt_template = PromptTemplate(template=summary_prompt, input_variables=["text"])
        summary_prompt_length = self.summary_llm_length_function(
            summary_prompt_template.format(**{input_var: "" for input_var in summary_prompt_template.input_variables})
        )
        # - 50 for safety
        self.summarize_chain_max_stuff_tokens = (
                self.summary_llm_max_context_size - summary_prompt_length - self.summary_max_length - 50
        )
        self.summarize_chain = load_summarize_chain(
            llm=self.summary_llm,
            prompt=summary_prompt_template,
            chain_type="stuff"
        )
        # "\n\n" is used for joining docs in stuff document chain
        self.join_length = self.summary_llm_length_function("\n\n")

    def total_length(self, sections):
        return (
            sum([self.summary_llm_length_function(section) for section in sections]) +
            self.join_length * (len(sections) - 1)
        )

    def summarize(self, sections):
        while self.total_length(sections) > self.summary_max_length:
            summaries = []
            docs = []
            count = 0
            for section in sections:
                section_length = self.summary_llm_length_function(section)
                count += section_length + self.join_length
                if count > self.summarize_chain_max_stuff_tokens:
                    summary = self.summarize_chain.run(docs).strip("\n")
                    summaries.append(summary)
                    count = section_length
                    docs = [Document(page_content=section)]
                else:
                    docs.append(Document(page_content=section))
            summary = self.summarize_chain.run(docs).strip("\n")
            summaries.append(summary)
            sections = summaries
        return "\n\n".join(sections)



import logging

import tiktoken
from langchain.text_splitter import RecursiveCharacterTextSplitter

from langsearch.pipelines.base import BasePipeline
from langsearch.utils import openai_length_function

logger = logging.getLogger(__name__)


class TextSplitterPipeline(BasePipeline):
    INPUTS = {
        "text": "text",
        "url": "url"
    }
    SECTIONS = "text_splitter_pipeline_sections"
    TEXT_SPLITTER_CLASS = RecursiveCharacterTextSplitter
    TEXT_SPLITTER_CLASS_PARAMS = {
        "chunk_size": 512,
        "chunk_overlap": 0,
        "length_function": openai_length_function
    }
    SIZE_CUTOFF = 40

    def __init__(self, text_splitter_class, text_splitter_class_params, size_cutoff, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text_splitter = text_splitter_class(**text_splitter_class_params)
        self.size_cutoff = size_cutoff

    @classmethod
    def from_crawler(cls, crawler):
        text_splitter_class = cls.get_setting_from_partial_key(crawler.settings, "TEXT_SPLITTER_CLASS")
        if isinstance(text_splitter_class, str):
            text_splitter_class = cls.get_from_dotted(text_splitter_class)
        text_splitter_class_params = cls.get_setting_from_partial_key(crawler.settings, "TEXT_SPLITTER_CLASS_PARAMS")
        if isinstance(text_splitter_class_params, str):
            text_splitter_class_params = cls.get_params_from_file(text_splitter_class_params)
        size_cutoff = cls.get_setting_from_partial_key(crawler.settings, "SIZE_CUTOFF")
        return cls(text_splitter_class, text_splitter_class_params, size_cutoff)

    def apply(self, item, spider):
        if not hasattr(self, "text") or self.text is None:
            return item
        if not hasattr(self, "url"):
            return item
        # TODO: We calculate the token count twice here. We should probably improve the text splitter so that it outputs
        # the token count along with the sections.
        sections = self.text_splitter.split_text(self.text)
        sections = [section for section in sections if self.text_splitter._length_function(section) > self.size_cutoff]
        item[self.SECTIONS] = sections
        return item



import scrapy
from scrapy.spiders import CrawlSpider, Rule
from scrapy.linkextractors import LinkExtractor

from langsearch.exceptions import SettingsError


class WebSpider(CrawlSpider):
    name = "langsearch_webspider"

    def __init__(self, start_urls, link_extractor_allow, link_extractor_deny, link_extractor_extra_args,
                 *args, **kwargs
                 ):
        if not hasattr(self, "start_urls") and start_urls is not None:
            self.start_urls = start_urls
        if len(link_extractor_allow) > 0 or len(link_extractor_deny) > 0:
            self.rules += (self.get_rule(link_extractor_allow, link_extractor_deny, link_extractor_extra_args),)
        super().__init__(*args, **kwargs)

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        start_urls = crawler.settings.get("LANGSEARCH_WEB_SPIDER_START_URLS")
        if start_urls is not None and not isinstance(start_urls, (list, tuple)):
            raise SettingsError(
                f"setting LANGSEARCH_WEB_SPIDER_START_URLS must be a list or tuple, got {type(start_urls)}."
            )
        link_extractor_allow = crawler.settings.get("LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_ALLOW", ())
        if not isinstance(link_extractor_allow, (list, tuple)):
            raise SettingsError(
                "setting LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_ALLOW must be a list or tuple, "
                f"got {type(link_extractor_allow)}"
            )

        link_extractor_deny = crawler.settings.get("LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_DENY", ())
        if not isinstance(link_extractor_deny, (list, tuple)):
            raise SettingsError(
                "setting LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_DENY must be a list or tuple, "
                f"got {type(link_extractor_deny)}"
            )

        link_extractor_extra_args = crawler.settings.get("LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_EXTRA_ARGS", {})
        if not isinstance(link_extractor_extra_args, dict):
            raise SettingsError(
                "setting LANGSEARCH_WEB_SPIDER_LINK_EXTRACTOR_EXTRA_ARGS must be a dict, "
                f"got {type(link_extractor_extra_args)}"
            )

        spider = super().from_crawler(crawler, start_urls, link_extractor_allow, link_extractor_deny,
                                      link_extractor_extra_args, *args, **kwargs
                                      )
        return spider

    def get_rule(self, link_extractor_allow, link_extractor_deny, link_extractor_extra_args):
        link_extractor = LinkExtractor(
            **{**link_extractor_extra_args, "allow": link_extractor_allow, "deny": link_extractor_deny}
            )
        return Rule(link_extractor, callback=self.parse, follow=True)

    def parse(self, response, **kwargs):
        item = {"response": response}
        return item

    # Scrapy's CrawlSpider will call this parser for the start_urls
    def parse_start_url(self, response, **kwargs):
        return self.parse(response, **kwargs)

    # We use dont_filter=False option for start_urls. Otherwise, the dupefilter doesn't store the start_url, leading
    # to two allowed requests to the same url.
    def start_requests(self):
        for url in self.start_urls:
            yield scrapy.Request(url)



from langsearch.pipelines.common.index import SimpleIndexPipeline
from langsearch.pipelines.common.storeitem import StoreItemPipeline
from langsearch.pipelines.common.textsplitter import TextSplitterPipeline
from langsearch.pipelines.common.inscriptis import InscriptisPipeline
from langsearch.pipelines.types.enumerations import ItemType


class GenericPlainTextPipeline:
    """
    Built-in pipelines should always use priorities between 400 and 600.
    """
    ITEM_TYPE = ItemType.TEXT

    ITEM_PIPELINES = {
        TextSplitterPipeline: 400,
        StoreItemPipeline: 410,
        SimpleIndexPipeline: 420
    }

    TEXT_SPLITTER_PIPELINE_INPUTS = {
        "text": lambda item: getattr(item["response"], "text"),
        "url": lambda item: getattr(item["response"], "url")
    }

    STORE_ITEM_PIPELINE_INPUTS = {
        "text": InscriptisPipeline.EXTRACTED_TEXT,
        "url": lambda item: getattr(item["response"], "url")
    }

    SIMPLE_INDEX_PIPELINE_INPUTS = {
        "sections": TextSplitterPipeline.SECTIONS,
        "url": lambda item: getattr(item["response"], "url"),
        "changed": StoreItemPipeline.CHANGED
    }

    PIPELINE_INPUTS = {
        TextSplitterPipeline: TEXT_SPLITTER_PIPELINE_INPUTS,
        StoreItemPipeline: STORE_ITEM_PIPELINE_INPUTS,
        SimpleIndexPipeline: SIMPLE_INDEX_PIPELINE_INPUTS
    }



import logging

from scrapy.exceptions import DropItem
from inscriptis import get_text, ParserConfig
from inscriptis.css_profiles import CSS_PROFILES

from langsearch.exceptions import SettingsError
from langsearch.pipelines.base import BasePipeline


logger = logging.getLogger(__name__)


class InscriptisPipeline(BasePipeline):
    INPUTS = {
        "html": "html",
        "url": "url"
    }
    EXTRACTED_TEXT = "inscriptis_pipeline_extracted_text"
    PARSER_CONFIG = ParserConfig(css=CSS_PROFILES["strict"].copy())

    def __init__(self, parser_config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parser_config = parser_config

    @classmethod
    def from_crawler(cls, crawler):
        parser_config = cls.get_setting_from_partial_key(crawler.settings, "PARSER_CONFIG")
        if not isinstance(parser_config, ParserConfig):
            raise SettingsError(
                f"setting with partial key PARSER_CONFIG of class {cls} must be a inscriptis.ParserConfig, "
                f"got {type(parser_config)}"
            )
        return cls(parser_config)

    def apply(self, item, spider):
        if not hasattr(self, "html") or self.html is None:
            return item
        if not hasattr(self, "url"):
            return item
        try:
            extracted_text = get_text(self.html, self.parser_config)
        except:
            message = f"Inscriptis failed to extract text for url {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            if isinstance(extracted_text, str):
                item[self.EXTRACTED_TEXT] = extracted_text
                return item
            else:
                message = f"Inscriptis extraction returned a non-string for url {self.url}"
                logger.warning(message)
                raise DropItem(message)



from io import BytesIO
import logging

from PIL import Image
from scrapy.exceptions import DropItem

from langsearch.exceptions import SettingsError
from langsearch.pipelines.base import BasePipeline


logger = logging.getLogger(__name__)


class ResizeImagePipeline(BasePipeline):
    INPUTS = {
        "body": "body",
        "url": "url",
    }
    RESIZED_BYTES = "resize_image_pipeline_resized_bytes"
    SCALE = 1
    RESIZE_ARGUMENTS = {}

    def __init__(self, scale, resized_width, resize_arguments, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale = scale
        self.resized_width = resized_width
        self.resize_arguments = resize_arguments

    @classmethod
    def from_crawler(cls, crawler):
        scale = cls.get_setting_from_partial_key(crawler.settings, "SCALE")
        try:
            resized_width = cls.get_setting_from_partial_key(crawler.settings, "RESIZED_WIDTH")
        except SettingsError:
            resized_width = None
        resize_arguments = cls.get_setting_from_partial_key(crawler.settings, "RESIZE_ARGUMENTS")
        return cls(scale, resized_width, resize_arguments)

    def get_resized_size(self, original_size):
        if self.resized_width is not None:
            return self.resized_width, int(self.resized_width * original_size[1] / original_size[0])
        else:
            return tuple(int(x * self.scale) for x in original_size)

    def apply(self, item, spider):
        if not hasattr(self, "body") or self.body is None:
            return item
        if not hasattr(self, "url"):
            return item
        try:
            with BytesIO(self.body) as original:
                im = Image.open(original)
                im_format = im.format
                resized_size = self.get_resized_size(im.size)
                im = im.resize(resized_size, **self.resize_arguments)
                with BytesIO() as resized:
                    im.save(resized, format=im_format)
                    item[self.RESIZED_BYTES] = resized.getvalue()
            return item
        except:
            message = f"Pillow failed to resize image for url {self.url}"
            logger.exception(message)
            raise DropItem(message)



from io import StringIO
import logging

import ffmpeg
import numpy as np
from scrapy.exceptions import DropItem
import whisper
from whisper.tokenizer import LANGUAGES
from whisper.utils import get_writer

from langsearch.exceptions import SettingsError
from langsearch.pipelines.base import BasePipeline


logger = logging.getLogger(__name__)


class WhisperPipeline(BasePipeline):
    INPUTS = {
        "body": "body",
        "url": "url"
    }
    TRANSCRIPTION = "whisper_pipeline_transcription"
    MODEL = "tiny"
    TRANSCRIPTION_OPTIONS = {}
    ALLOWED_LANGUAGES = list(LANGUAGES.keys())
    OUTPUT_FORMAT = "txt"

    def __init__(self, model, transcription_options, allowed_languages, output_format, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = whisper.load_model(model)
        self.transcription_options = transcription_options
        self.allowed_languages = allowed_languages
        self.writer = get_writer(output_format, output_dir="")  # We don't use the output dir

    @classmethod
    def from_crawler(cls, crawler):
        model = cls.get_setting_from_partial_key(crawler.settings, "MODEL")
        if not isinstance(model, str):
            raise SettingsError(
                f"setting with partial key MODEL of class {cls} must be a str, "
                f"got {type(model)}"
            )
        transcription_options = cls.get_setting_from_partial_key(crawler.settings, "TRANSCRIPTION_OPTIONS")
        if not isinstance(transcription_options, dict):
            raise SettingsError(
                f"setting with partial key TRANSCRIPTION_OPTIONS of class {cls} must be a dict, "
                f"got {type(transcription_options)}"
            )
        allowed_languages = cls.get_setting_from_partial_key(crawler.settings, "ALLOWED_LANGUAGES")
        if not isinstance(allowed_languages, list):
            raise SettingsError(
                f"setting with partial key ALLOWED_LANGUAGES of class {cls} must be a list, "
                f"got {type(allowed_languages)}"
            )
        output_format = cls.get_setting_from_partial_key(crawler.settings, "OUTPUT_FORMAT")
        allowed_output_formats = ["txt", "vtt", "srt", "tsv"]
        if output_format not in allowed_output_formats:
            raise SettingsError(
                f"setting with partial key OUTPUT_FORMAT of class {cls} must be one of {allowed_output_formats}, "
                f"got {output_format}"
            )
        return cls(model, transcription_options, allowed_languages, output_format)

    def load_audio(self):
        try:
            out, _ = (
                ffmpeg.input("pipe:", threads=0)
                .output("-", format="s16le", acodec="pcm_s16le", ac=1, ar=16000)  # sample rate is hardcoded in whisper
                .run(cmd="ffmpeg", capture_stdout=True, capture_stderr=True, input=self.body)
            )
        except ffmpeg.Error as e:
            raise RuntimeError(f"Failed to load audio: {e.stderr.decode()}") from e

        return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0

    def apply(self, item, spider):
        if not hasattr(self, "body") or self.body is None:
            return item
        if not hasattr(self, "url"):
            return item
        try:
            audio = self.load_audio()
            audio_for_lang_detection = whisper.pad_or_trim(audio)
            mel = whisper.log_mel_spectrogram(audio_for_lang_detection).to(self.model.device)
            # TODO: We are doing language detection twice, once here and once in transcribe(). Make this more optimal.
            _, probs = self.model.detect_language(mel)
            detected_lang = max(probs, key=probs.get)
            if detected_lang not in self.allowed_languages:
                logger.info(f"Detected language {detected_lang} of audio at URL {self.url} "
                            f"is not in allowed languages {self.allowed_languages}"
                            )
                return item
            result = whisper.transcribe(self.model, audio, **self.transcription_options)
            text_stream = StringIO()
            self.writer.write_result(result, file=text_stream)
            item[self.TRANSCRIPTION] = text_stream.getvalue()
            return item
        except:
            message = f"Whisper failed to extract text for url {self.url}"
            logger.exception(message)
            raise DropItem(message)



import importlib
import json
import logging
from pathlib import Path
import re

from langsearch.exceptions import SettingsError


logger = logging.getLogger(__name__)


# TODO: Use ItemAdapter in all pipelines
class BasePipeline:
    INPUTS = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def get_setting_from_partial_key(cls, settings_obj, partial_key, default_class_var=None):
        if default_class_var is None:
            default_class_var = partial_key
        logger.info(f"Getting settings for partial key {partial_key} for class {cls}")
        for this_cls in cls.mro():
            key = f"LANGSEARCH_{this_cls.__name__.upper()}_{partial_key}"
            logger.debug(f"Trying environment variable {key}")
            value = settings_obj.get(key)
            if value is None:
                logger.debug(f"Trying class variable {default_class_var} of class {this_cls}")
                if hasattr(this_cls, default_class_var):
                    value = getattr(this_cls, default_class_var)
                    logger.info(f"Resolved partial key {partial_key} to class var {this_cls}.{default_class_var} "
                                f"with value {value}"
                                )
                    return value
            else:
                logger.info(f"Resolved partial key {partial_key} to env var {key} with value {value}")
                return value
        raise SettingsError(f"Failed to find value for setting with partial key {partial_key} for class {cls}")

    @staticmethod
    def get_from_dotted(dotted):
        parts = re.split(r"\.", dotted)
        module = importlib.import_module(".".join(parts[:-1]))
        return getattr(module, parts[-1])

    @staticmethod
    def get_params_from_file(filepath):
        path = Path(filepath)
        if not path.exists():
            raise SettingsError(f"Path {filepath} does not exist")
        if not path.is_file():
            raise SettingsError(f"Path {filepath} is not a file")
        with path.open("r") as f:
            params = json.load(f)
        return params

    def process_item(self, item, spider):
        self.prepare_inputs(item)
        item = self.apply(item, spider)
        return item

    def prepare_inputs(self, item):
        for argument, item_type_and_key in self.INPUTS.items():
            if not isinstance(item_type_and_key, dict):
                key = item_type_and_key
            else:
                try:
                    key = item_type_and_key[item["type"]]
                except KeyError:
                    setattr(self, argument, None)
                    continue
            if callable(key):
                try:
                    value = key(item)
                except:
                    setattr(self, argument, None)
                else:
                    setattr(self, argument, value)
                continue
            try:
                value = item[key]
            except KeyError:
                setattr(self, argument, None)
            else:
                setattr(self, argument, value)

    def apply(self, item, spider):
        """
        Should check if it has the right instance variables set to carry out the processing. Otherwise, DropItem.
        """
        raise NotImplementedError



from pathlib import Path

import scrapy
from scrapy.spiders import Spider

from langsearch.exceptions import SettingsError
from langsearch.utils import get_regex_from_list


class FileSpider(Spider):
    name = "langsearch_filespider"

    def __init__(self, start_folders, follow_subfolders, follow_symlinks, allow, deny, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_folders = []
        for folder in start_folders:
            folder = Path(folder)
            if not folder.is_absolute():
                raise SettingsError(f"folder {folder} is not an absolute path")
            if not folder.exists():
                raise SettingsError(f"folder {folder} does not exist")
            if not folder.is_dir():
                raise SettingsError(f"{folder} is not a folder")
            self.start_folders.append(folder)

        self.follow_subfolders = follow_subfolders
        self.follow_symlinks = follow_symlinks
        self.allow = get_regex_from_list(allow) if len(allow) > 0 else None
        self.deny = get_regex_from_list(deny) if len(deny) > 0 else None

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        start_folders = crawler.settings.get("LANGSEARCH_FILE_SPIDER_START_FOLDERS")
        if start_folders is None:
            raise SettingsError("setting LANGSEARCH_FILE_SPIDER_START_FOLDERS is missing")
        if not isinstance(start_folders, (list, tuple)):
            raise SettingsError(
                f"setting LANGSEARCH_FILESPIDER_START_FOLDERS must be a list or tuple, got {type(start_folders)}."
            )

        follow_subfolders = crawler.settings.get("LANGSEARCH_FILE_SPIDER_FOLLOW_SUBFOLDERS", False)
        if not isinstance(follow_subfolders, bool):
            raise SettingsError(
                f"setting LANGSEARCH_FILE_SPIDER_FOLLOW_SUBFOLDERS must be a boolean, got {type(follow_subfolders)}"
            )

        follow_symlinks = crawler.settings.get("LANGSEARCH_FILE_SPIDER_FOLLOW_SYMLINKS", False)
        if not isinstance(follow_symlinks, bool):
            raise SettingsError(
                f"setting LANGSEARCH_FILE_SPIDER_FOLLOW_SYMLINKS must be a boolean, got {type(follow_symlinks)}"
            )

        allow = crawler.settings.get("LANGSEARCH_FILE_SPIDER_ALLOW", ())
        if not isinstance(allow, (list, tuple)):
            raise SettingsError(f"setting LANGSEARCH_FILE_SPIDER_ALLOW must be a list or tuple, got {type(allow)}")

        deny = crawler.settings.get("LANGSEARCH_FILE_SPIDER_DENY", ())
        if not isinstance(allow, (list, tuple)):
            raise SettingsError(f"setting LANGSEARCH_FILE_SPIDER_DENY must be a list or tuple, got {type(deny)}")

        spider = super().from_crawler(crawler, start_folders, follow_subfolders, follow_symlinks, allow, deny,
                                      *args, **kwargs
                                      )
        return spider

    def filtered(self, path):
        if self.deny is not None and self.deny.search(str(path)) is not None:
            return True
        if self.allow is not None and self.allow.search(str(path)) is None:
            return True
        return False

    def get_files(self, folder):
        for child in folder.iterdir():
            if child.is_symlink():
                if self.follow_symlinks:
                    child = child.resolve()
                    if self.filtered(child):
                        continue
                else:
                    continue
            if child.is_dir():
                if self.follow_subfolders:
                    yield from self.get_files(child)
            elif not self.filtered(child):
                yield child

    def start_requests(self):
        for folder in self.start_folders:
            for file in self.get_files(folder):
                yield scrapy.Request(file.as_uri(), callback=self.parse)

    def parse(self, response, **kwargs):
        item = {"response": response}
        return item



import logging

from lxml import etree
from scrapy.exceptions import DropItem

from langsearch.pipelines.base import BasePipeline

logger = logging.getLogger(__name__)


class FixHTMLPipeline(BasePipeline):
    INPUTS = {
        "html": "html",
        "url": "url"
    }
    FIXED_HTML = "fix_html_pipeline_html"

    def apply(self, item, spider):
        if not hasattr(self, "html") or self.html is None:
            return item
        try:
            parsed = etree.HTML(self.html)
        except:
            message = f"Failed to parse response html for url {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            item[self.FIXED_HTML] = etree.tostring(parsed, method="html", encoding=str)
            return item




import logging

from langchain.chains.base import Chain
from langchain.chains.hyde.prompts import PROMPT_MAP
from langchain.chains.llm import LLMChain
from langchain.llms import OpenAIChat
from pydantic import Extra

from langsearch.chains.qa import QAChain

logger = logging.getLogger(__name__)


class HYDEChain(Chain):
    class Config:
        extra = Extra.allow

    def __init__(self,
                 *args,
                 hyde_llm_chain=LLMChain(
                     llm=OpenAIChat(temperature=0.7),
                     prompt=PROMPT_MAP["web_search"]
                 ),
                 hyde_llm_chain_question_input_name="QUESTION",
                 langsearch_qa_chain=QAChain(
                     document_search_question_input_name="hyde_llm_output"
                 ),
                 **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.hyde_llm_chain = hyde_llm_chain
        self.hyde_llm_chain_question_input_name = hyde_llm_chain_question_input_name
        self.langsearch_qa_chain = langsearch_qa_chain

    @property
    def input_keys(self):
        return self.hyde_llm_chain.input_keys

    @property
    def output_keys(self):
        return self.langsearch_qa_chain.output_keys

    def _call(self, inputs):
        outputs = self.hyde_llm_chain(inputs, return_only_outputs=True)
        logger.debug("Hyde LLM chain outputs: %s", outputs)
        return self.langsearch_qa_chain(
            {
                self.langsearch_qa_chain.document_search_question_input_name: outputs[self.hyde_llm_chain.output_key],
                self.langsearch_qa_chain.qa_chain_question_input_name: inputs[self.hyde_llm_chain_question_input_name]
            },
            return_only_outputs=True
        )



import logging

from langchain.chains.base import Chain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.question_answering import load_qa_chain
from langchain.llms import OpenAIChat
from pydantic import Extra

from langsearch.pipelines.common.index import SimpleIndexPipeline
from langsearch.utils import openai_length_function

logger = logging.getLogger(__name__)


class QAChain(Chain):
    class Config:
        extra = Extra.allow

    def __init__(self,
                 *args,
                 qa_chain=load_qa_chain(llm=OpenAIChat(temperature=0)),
                 qa_chain_question_input_name="question",
                 document_search_question_input_name="question",
                 length_function=openai_length_function,
                 max_context_size=4096,
                 method_or_docs=SimpleIndexPipeline().get_similar_sections,
                 method_args=None,
                 **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.qa_chain = qa_chain
        self.qa_chain_question_input_name = qa_chain_question_input_name
        self.document_search_question_input_name = document_search_question_input_name
        self.length_function = length_function
        self.max_context_size = max_context_size
        self.method_or_docs = method_or_docs
        if method_args is None:
            method_args = {"limit": 5}
        self.method_args = method_args

    @property
    def input_keys(self):
        return list({self.qa_chain_question_input_name, self.document_search_question_input_name})

    @property
    def output_keys(self):
        return self.qa_chain.output_keys + ["docs"]

    def _call(self, inputs):
        if callable(self.method_or_docs):
            docs = self.method_or_docs(inputs[self.document_search_question_input_name], **self.method_args)
        else:
            docs = self.method_or_docs
        if isinstance(self.qa_chain, StuffDocumentsChain):
            prompt_template = self.qa_chain.llm_chain.prompt
            empty_prompt = prompt_template.format(**{input_var: "" for input_var in prompt_template.input_variables})
            prompt_length = self.length_function(empty_prompt)
            question_length = self.length_function(inputs[self.qa_chain_question_input_name])
            # -256 minimum for answer, -100 for safety
            remaining = self.max_context_size - prompt_length - question_length - 256 - 100

            token_count = 0
            trimmed_docs = []
            for doc in docs:
                token_count += self.length_function(doc.page_content) + self.length_function("\n\n")
                if token_count > remaining:
                    break
                trimmed_docs.append(doc)
            docs = trimmed_docs
        # TODO: how to pass max_tokens in map_reduce_chain?
        qa_chain_inputs = inputs.copy()
        if self.document_search_question_input_name != self.qa_chain_question_input_name:
            del qa_chain_inputs[self.document_search_question_input_name]
        return {"docs": docs,
                **self.qa_chain({**qa_chain_inputs, self.qa_chain.input_key: docs}, return_only_outputs=True)
                }



import datetime
from difflib import SequenceMatcher
import logging
import os

from scrapy.exceptions import DropItem

from langsearch.pipelines.base import BasePipeline
from langsearch.pipelines.common.mixins.weaviatedb import WeaviateMixin

logger = logging.getLogger(__name__)


class BaseStoreItemPipeline(BasePipeline):
    INPUTS = {
        "text": "text",
        "url": "url",
    }
    CHANGED = "store_item_pipeline_changed"
    CLASS_SCHEMA = {
        "class": "CrawlData",
        "vectorizer": "none",
        "vectorIndexConfig": {
            "skip": True
        },
        "invertedIndexConfig": {
            "indexNullState": True
        },
        "properties": [
            {
                "name": "url",
                "description": "The URL containing the section",
                "dataType": ["string"],
            },
            {
                "name": "text",
                "description": "Text of the document, if available",
                "dataType": ["text"],
            },
            {
                "name": "last_seen",
                "description": "When this URL was last seen in the crawling process",
                "dataType": ["date"],
            },
        ],
    }
    DUPLICATE_CUTOFF = 95

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        class_schema = self.__class__.get_setting_from_partial_key(os.environ, "CLASS_SCHEMA")
        if isinstance(class_schema, str):
            class_schema = self.get_params_from_file(class_schema)
        self.class_schema = class_schema
        self.class_name = self.class_schema["class"]
        if not self.weaviate.class_exists(self.class_name):
            self.weaviate.create_class(self.class_schema)
        self.duplicate_cutoff = self.__class__.get_setting_from_partial_key(os.environ, "DUPLICATE_CUTOFF")

    @staticmethod
    def similarity(text1, text2):
        m = SequenceMatcher(None, text1, text2)
        return m.real_quick_ratio() * 100

    def get_all_data_obj_for_url(self):
        where_filter = {
            "path": ["url"],
            "operator": "Equal",
            "valueString": self.url
        }
        result = (
            self.weaviate.client.query
            .get(self.class_name, ["text"])
            .with_additional(["id"])
            .with_where(where_filter)
            .do()
        )
        return result["data"]["Get"][self.class_name]

    def apply(self, item, spider):
        if not hasattr(self, "url") or self.url is None:
            return item
        try:
            data = self.get_all_data_obj_for_url()
            count = len(data)
            if count > 1:
                message = f"Found {count} data obj for unique property URL {self.url} in class {self.class_name}"
                logger.warning(message)
                raise RuntimeError(message)
            elif count == 0:
                self.weaviate.client.data_object.create(
                    class_name=self.class_name,
                    data_object={
                        "url": self.url,
                        "text": self.text if hasattr(self, "text") else None,
                        "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                    }
                )
                item[self.CHANGED] = True
                return item
            else:
                unique_id = data[0]["_additional"]["id"]
                if hasattr(self, "text"):
                    text = data[0]["text"]
                    if self.similarity(text, self.text) <= self.duplicate_cutoff:
                        self.weaviate.client.data_object.update(
                            class_name=self.class_name,
                            uuid=unique_id,
                            data_object={
                                "url": self.url,
                                "text": self.text,
                                "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                            }
                        )
                        item[self.CHANGED] = True
                        return item
                self.weaviate.client.data_object.update(
                    class_name=self.class_name,
                    uuid=unique_id,
                    data_object={
                        "last_seen": datetime.datetime.now(datetime.timezone.utc).isoformat()
                    }
                )
                item[self.CHANGED] = False
                return item
        except:
            message = f"Error while processing item with URL {self.url}"
            logger.exception(message)
            raise DropItem(message)


class StoreItemPipeline(BaseStoreItemPipeline, WeaviateMixin):
    pass



import logging

from scrapy.exceptions import DropItem
from tika import detector

from langsearch.pipelines.types.enumerations import ItemType

logger = logging.getLogger(__name__)


class DetectItemTypePipeline:
    def process_item(self, item, spider):
        response = item["response"]
        try:
            content_type = response.headers["Content-Type"]
            if isinstance(content_type, bytes):
                content_type = content_type.decode()
            if content_type.startswith("text/html"):
                item["type"] = ItemType.HTML
                return item
        except KeyError:
            pass
        mime_type = detector.from_buffer(response.body)
        if mime_type == "text/html":
            item["type"] = ItemType.HTML
            return item
        elif mime_type.startswith("text"):
            item["type"] = ItemType.TEXT
            return item
        elif mime_type.startswith(("audio", "video")):
            item["type"] = ItemType.AUDIO
            return item
        elif mime_type.startswith("image"):
            item["type"] = ItemType.IMAGE
            return item
        else:
            item["type"] = ItemType.OTHER
            return item
        message = f"Item with url {response.url} could not be classified into a type"
        logger.warning(message)
        raise DropItem(message)



import logging
from scrapy.http import Response
from scrapy.utils.httpobj import urlparse_cached

from langsearch.exceptions import IgnoreResponse
from langsearch.utils import get_regex_from_list

logger = logging.getLogger(__name__)


class RegexFilterMiddleware:
    def __init__(self, allow, deny, stats):
        self.allow = get_regex_from_list(allow) if len(allow) > 0 else None
        self.deny = get_regex_from_list(deny) if len(deny) > 0 else None
        self.stats = stats

    @classmethod
    def from_crawler(cls, crawler):
        allow = crawler.settings.get("LANGSEARCH_REGEX_FILTER_MIDDLEWARE_ALLOW", ())
        deny = crawler.settings.get("LANGSEARCH_REGEX_FILTER_MIDDLEWARE_DENY", ())
        return cls(allow, deny, crawler.stats)

    def filtered(self, response):
        parsed = urlparse_cached(response)
        full_path = parsed.netloc + parsed._replace(scheme="", netloc="").geturl()
        if self.deny is not None and self.deny.search(full_path) is not None:
            return True
        if self.allow is not None and self.allow.search(full_path) is None:
            return True
        return False

    def process_spider_input(self, response, spider):
        if not isinstance(response, Response):
            return
        if self.filtered(response):
            message = f"{self.__class__.__name__} has filtered response with URL {response.url}"
            logger.debug(message)
            self.stats.inc_value(f"{self.__class__.__name__}/filtered")
            raise IgnoreResponse(message)

    # This seems to be called only on start_urls, because the requests from the CrawlSpider rule have an errback, which
    # is called instead
    def process_spider_exception(self, response, exception, spider):
        if isinstance(exception, IgnoreResponse):
            return []



import logging

from scrapy.exceptions import DropItem
from tika import parser

from langsearch.pipelines.base import BasePipeline

logger = logging.getLogger(__name__)


class TikaPipeline(BasePipeline):
    INPUTS = {
        "body": "body",
        "url": "url"
    }
    XML_OUTPUT = "tika_pipeline_xml_output"

    def apply(self, item, spider):
        if not hasattr(self, "body") or self.body is None:
            return item
        if not hasattr(self, "url"):
            return item
        try:
            # TODO: Add timeout for docker compose like setup.
            xml_output = parser.from_buffer(self.body, xmlContent=True)["content"]
        except:
            message = f"Tika failed to extract text for url {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            if isinstance(xml_output, str):
                item[self.XML_OUTPUT] = xml_output
                return item
            else:
                message = f"Tika extraction returned a non-string for url {self.url}"
                logger.info(message)
                raise DropItem(message)



import logging

from lxml import etree
from readability import Document
from scrapy.exceptions import DropItem

from langsearch.pipelines.base import BasePipeline


logger = logging.getLogger(__name__)


class PythonReadabilityPipeline(BasePipeline):
    INPUTS = {
        "html": "html",
        "url": "url"
    }
    HTML_WITHOUT_BP = "python_readability_pipeline_text_without_bp"

    def apply(self, item, spider):
        if not hasattr(self, "html") or self.html is None:
            return item
        if not hasattr(self, "url"):
            return item
        try:
            html_without_bp = Document(self.html).summary()
        except:
            message = f"Python-readability failed to remove boilerplate for url {self.url}"
            logger.exception(message)
            raise DropItem(message)
        else:
            try:
                parsed = etree.HTML(html_without_bp)
            except:
                message = f"Python-readability did not return valid HTML for {self.url}"
                logger.exception(message)
                raise DropItem(message)
            else:
                item[self.HTML_WITHOUT_BP] = html_without_bp
                return item



import datetime
import logging
import os
import time

from requests.exceptions import ConnectionError, HTTPError
from weaviate import Client
from weaviate.util import image_encoder_b64


logger = logging.getLogger(__name__)


class WeaviateDB:
    def __init__(self, base_url):
        # TODO: Added timeout for docker compose like setup. Look for a better solution.
        # TODO: Use `startup_period` argument of `weaviate.Client` when it becomes available.
        for _i in range(20):
            try:
                self.client = Client(base_url)
            except (ConnectionError, HTTPError):
                time.sleep(1)
            else:
                break

    def class_exists(self, class_name):
        schema = self.client.schema.get()
        classes = [c["class"] for c in schema["classes"]]
        return class_name in classes

    def create_class(self, class_schema):
        self.client.schema.create_class(class_schema)

    def add(self, class_name, data, batch_size=5):
        self.client.batch.configure(batch_size=batch_size)
        with self.client.batch as batch:
            for data_object in data:
                batch.add_data_object(
                    class_name=class_name,
                    data_object=data_object
                )

    def update_property_with_current_datetime(self, class_name, where_filter, property_name):
        result = (
            self.client.query
            .get(class_name)
            .with_additional(["id"])
            .with_where(where_filter)
            .do()
        )
        for element in result["data"]["Get"][class_name]:
            unique_id = element["_additional"]["id"]
            self.client.data_object.update(
                class_name=class_name,
                uuid=unique_id,
                data_object={
                    property_name: datetime.datetime.now(datetime.timezone.utc).isoformat()
                }
            )

    def get_similar(self, class_name, similar_to, query_attrs, limit=4, where_filter=None,
                    distance=None, hybrid=False, alpha=0.75
                    ):
        near_text = {"concepts": [similar_to]}
        result = (
            self.client.query
            .get(class_name, query_attrs)
        )
        if where_filter is not None:
            result = result.with_where(where_filter)
        if hybrid:
            result = result.with_hybrid(similar_to, alpha=alpha)
        else:
            if distance is not None:
                near_text["distance"] = distance
            result = result.with_near_text(near_text)
        result = result.with_limit(limit).do()
        logger.debug("Weaviate query result: %s", result)
        return result["data"]["Get"][class_name]

    def get_near_image(self, class_name, similar_to, query_attrs, limit):
        near_image = {"image": image_encoder_b64(similar_to)}
        result = (
            self.client.query
            .get(class_name, query_attrs)
            .with_near_image(near_image)
            .with_limit(limit)
            .do()
        )
        return result["data"]["Get"][class_name]


class WeaviateMixin:
    WEAVIATE_BASE_URL = "http://localhost:8080"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        weaviate_base_url = self.__class__.get_setting_from_partial_key(os.environ, "WEAVIATE_BASE_URL")
        self.weaviate = WeaviateDB(weaviate_base_url)
